// Import only the required methods from D3 to reduce bundle size when using modules.
import {
  select,
  json,
  forceSimulation,
  forceLink,
  forceManyBody,
  forceCollide,
  forceCenter,
  drag,
} from 'd3';

// Setup the SVG container.
const svg = select('svg');
const width = +svg.attr('width');
const height = +svg.attr('height');
const colorScale = 'Lavender';

// Configure SVG with responsive attributes.
svg
  .attr('preserveAspectRatio', 'xMidYMid meet')
  .attr('viewBox', `0 0 ${width} ${height}`);

// Marker definition moved closer to its usage.
const defs = svg.append('defs');
defs
  .append('marker')
  .attr('id', 'node')
  .attr('viewBox', '0 -2.5 5 5')
  .attr('refX', 15)
  .attr('refY', 0)
  .attr('markerWidth', 5)
  .attr('markerHeight', 6)
  .attr('orient', 'auto')
  .append('path')
  .attr('d', 'M0,-2.5L5,0L0,2.5')
  .attr('fill', 'black');

// Load data and setup graph.
json('graph.json').then((data) => {
  // Use structure cloning for performance, if necessary.
  const links = data.links.map((d) => ({ ...d }));
  const nodes = data.nodes.map((d) => ({ ...d }));

  // Setup forces.
  const simulation = forceSimulation(nodes)
    .force(
      'link',
      forceLink(links)
        .id((d) => d.id)
        .distance(40),
    )
    .force('charge', forceManyBody().strength(-500))
    .force('collision', forceCollide(105))
    .force('center', forceCenter(width / 2, height / 2));

  // Define drag behaviors.
  const dragBehavior = drag()
    .on('start', (event) => {
      if (!event.active)
        simulation.alphaTarget(0.5).restart();
      event.subject.fx = event.subject.x;
      event.subject.fy = event.subject.y;
    })
    .on('drag', (event) => {
      event.subject.fx = event.x;
      event.subject.fy = event.y;
    })
    .on('end', (event) => {
      if (!event.active) simulation.alphaTarget(0);
      event.subject.fx = null;
      event.subject.fy = null;
    });

  // Create links.
  svg
    .append('g')
    .selectAll('line')
    .data(links)
    .join('line')
    .attr('marker-end', 'url(#node)');

  // Create nodes.
  const node = svg
    .append('g')
    .selectAll('g')
    .data(nodes)
    .enter()
    .append('g');

  node
    .append('circle')
    .attr('r', 40)
    .attr('fill', colorScale)
    .call(dragBehavior)
    .on('mouseover', function () {
      select(this).classed('nodeHover', true);
    })
    .on('mouseout', function () {
      select(this).classed('nodeHover', false);
    });

  node
    .append('text')
    .text((d) => d.id)
    .attr('dy', '0.3em');

  // Tick function for updating positions.
  simulation.on('tick', () => {
    svg
      .selectAll('line')
      .attr('x1', (d) => d.source.x)
      .attr('y1', (d) => d.source.y)
      .attr('x2', (d) => d.target.x)
      .attr('y2', (d) => d.target.y);

    node.attr(
      'transform',
      (d) => `translate(${d.x}, ${d.y})`,
    );
  });
});
